@@ -1,5 +1,5 @@
 /* GNU ed - The GNU line editor.
-   Copyright (C) 2006-2025 Antonio Diaz Diaz.
+   Copyright (C) 2006-2026 Antonio Diaz Diaz.
 
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -47,13 +47,13 @@
 
 
 static const char * const program_name = "ed";
-static const char * const program_year = "2025";
+static const char * const program_year = "2026";
 static const char * invocation_name = "ed";		/* default value */
 
 static bool extended_regexp_ = false;	/* use EREs */
 static bool quiet = false;		/* suppress diagnostics */
 static bool restricted_ = false;	/* run in restricted mode */
-static bool safe_names = true;		/* reject chars 1-31 in file names */
+static bool safe_names = true;		/* reject control chars in file names */
 static bool scripted_ = false;		/* suppress byte counts and ! prompt */
 static bool strip_cr_ = false;		/* strip trailing CRs */
 static bool traditional_ = false;	/* be backwards compatible */
@@ -68,41 +68,43 @@ bool traditional( void ) { return traditional_; }
 
 static void show_help( void )
   {
-  printf( "GNU ed is a line-oriented text editor. It is used to create, display,\n"
-          "modify and otherwise manipulate text files, both interactively and via\n"
-          "shell scripts. A restricted version of ed, red, can only edit files in\n"
-          "the current directory and cannot execute shell commands. Ed is the\n"
-          "'standard' text editor in the sense that it is the original editor for\n"
-          "Unix, and thus widely available. For most purposes, however, it is\n"
-          "superseded by full-screen editors.\n"
-          "\nUsage: %s [options] [[+line] file]\n", invocation_name );
-  printf( "\nThe file name may be preceded by '+line', '+/RE', or '+?RE' to set the\n"
-          "current line to the line number specified or to the first or last line\n"
-          "matching the regular expression 'RE'.\n"
-          "\nThe environment variable LINES can be used to set the initial window size.\n"
-          "\nOptions:\n"
-          "  -h, --help                 display this help and exit\n"
-          "  -V, --version              output version information and exit\n"
-          "  -E, --extended-regexp      use extended regular expressions\n"
-          "  -G, --traditional          run in compatibility mode\n"
-          "  -l, --loose-exit-status    exit with 0 status even if a command fails\n"
-          "  -p, --prompt=STRING        use STRING as an interactive prompt\n"
-          "  -q, --quiet, --silent      suppress diagnostics written to stderr\n"
-          "  -r, --restricted           run in restricted mode\n"
-          "  -s, --script               suppress byte counts and '!' prompt\n"
-          "  -v, --verbose              be verbose; equivalent to the 'H' command\n"
-          "      --strip-trailing-cr    strip carriage returns at end of text lines\n"
-          "      --unsafe-names         allow control characters 1-31 in file names\n"
-          "\nStart edit by reading in 'file' if given.\n"
-          "If 'file' begins with a '!', read output of shell command.\n"
-          "\nExit status: 0 for a normal exit, 1 for environmental problems\n"
-          "(invalid command-line options, memory exhausted, command failed, etc),\n"
-          "2 for problems with the input file (file not found, buffer modified,\n"
-          "I/O errors), 3 for an internal consistency error (e.g., bug) which caused\n"
-          "ed to panic.\n"
-          "\nReport bugs to bug-ed@gnu.org\n"
-          "Ed home page: http://www.gnu.org/software/ed/ed.html\n"
-          "General help using GNU software: http://www.gnu.org/gethelp\n" );
+  fputs(
+    "GNU ed is a line-oriented text editor. It is used to create, display,\n"
+    "modify and otherwise manipulate text files, both interactively and via\n"
+    "shell scripts. A restricted version of ed, red, can only edit files in\n"
+    "the current directory and cannot execute shell commands. Ed is the\n"
+    "'standard' text editor in the sense that it is the original editor for\n"
+    "Unix, and thus widely available. For most purposes, however, it is\n"
+    "superseded by full-screen editors.\n", stdout );
+  printf( "\nUsage: %s [options] [[+line] file]\n", invocation_name );
+  fputs(
+    "\nThe file name may be preceded by '+line', '+/RE', or '+?RE' to set the\n"
+    "current line to the line number specified or to the first or last line\n"
+    "matching the regular expression 'RE'.\n"
+    "\nStart edit by reading in 'file' if given.\n"
+    "If 'file' begins with a '!', read output of shell command.\n"
+    "\nThe environment variable LINES can be used to set the initial window size.\n"
+    "\nOptions:\n"
+    "  -h, --help                 display this help and exit\n"
+    "  -V, --version              output version information and exit\n"
+    "  -E, --extended-regexp      use extended regular expressions\n"
+    "  -G, --traditional          run in compatibility mode\n"
+    "  -l, --loose-exit-status    exit with 0 status even if a command fails\n"
+    "  -p, --prompt=STRING        use STRING as an interactive prompt\n"
+    "  -q, --quiet, --silent      suppress diagnostics written to stderr\n"
+    "  -r, --restricted           run in restricted mode\n"
+    "  -s, --script               suppress byte counts and '!' prompt\n"
+    "  -v, --verbose              be verbose; equivalent to the 'H' command\n"
+    "      --strip-trailing-cr    strip carriage returns at end of text lines\n"
+    "      --unsafe-names         allow control characters in file names\n"
+    "\n*Exit status*\n"
+    "0 for a normal exit, 1 for environmental problems (invalid command-line\n"
+    "options, memory exhausted, command failed, etc), 2 for problems with the\n"
+    "input file (file not found, buffer modified, I/O errors), 3 for an internal\n"
+    "consistency error (e.g., bug) which caused ed to panic.\n"
+    "\nReport bugs to bug-ed@gnu.org\n"
+    "Ed home page: http://www.gnu.org/software/ed/ed.html\n"
+    "General help using GNU software: http://www.gnu.org/gethelp\n", stdout );
   }
 
 
@@ -111,23 +113,22 @@ static void show_version( void )
   printf( "GNU %s %s\n", program_name, PROGVERSION );
   printf( "Copyright (C) 1994 Andrew L. Moore.\n"
           "Copyright (C) %s Antonio Diaz Diaz.\n", program_year );
-  printf( "License GPLv2+: GNU GPL version 2 or later <http://gnu.org/licenses/gpl.html>\n"
-          "This is free software: you are free to change and redistribute it.\n"
-          "There is NO WARRANTY, to the extent permitted by law.\n" );
+  fputs( "License GPLv2+: GNU GPL version 2 or later <http://gnu.org/licenses/gpl.html>\n"
+         "This is free software: you are free to change and redistribute it.\n"
+         "There is NO WARRANTY, to the extent permitted by law.\n", stdout );
   }
 
 
-void print_filename( const char * const filename, const bool to_stdout )
+void print_escaped( const char * p, const bool to_stdout )
   {
   FILE * const fp = to_stdout ? stdout : stderr;
-  if( safe_names ) { fputs( filename, fp ); return; }
-  const char * p;
-  for( p = filename; *p; ++p )
+  for( ; *p; ++p )
     {
     const unsigned char ch = *p;
     if( ch == '\\' ) { putc( ch, fp ); putc( ch, fp ); continue; }
-    if( ch >= 32 ) { putc( ch, fp ); continue; }
+    if( ch >= 32 && ch <= 126 ) { putc( ch, fp ); continue; }
     putc( '\\', fp );
+    const char e = escchar( ch ); if( e ) { putc( e, fp ); continue; }
     putc( ( ( ch >> 6 ) & 7 ) + '0', fp );
     putc( ( ( ch >> 3 ) & 7 ) + '0', fp );
     putc( ( ch & 7 ) + '0', fp );
@@ -140,7 +141,7 @@ void show_warning( const char * const filename, const char * const msg )
   if( !quiet )
     {
     if( filename && filename[0] )
-      { print_filename( filename, false ); fputs( ": ", stderr ); }
+      { print_escaped( filename, false ); fputs( ": ", stderr ); }
     fprintf( stderr, "%s\n", msg );
     }
   }
@@ -151,7 +152,7 @@ void show_strerror( const char * const filename, const int errcode )
   if( !quiet )
     {
     if( filename && filename[0] )
-      { print_filename( filename, false ); fputs( ": ", stderr ); }
+      { print_escaped( filename, false ); fputs( ": ", stderr ); }
     fprintf( stderr, "%s\n", strerror( errcode ) );
     }
   }
@@ -195,7 +196,7 @@ bool may_access_filename( const char * const name )
   {
   const int len = strlen( name );
   if( len <= 0 || name[len-1] == '/' )
-    { set_error_msg( "Is a directory" ); return false; }
+    { set_error_msg( "Invalid filename" ); return false; }
   if( restricted_ )
     {
     if( name[0] == '!' )
@@ -203,13 +204,16 @@ bool may_access_filename( const char * const name )
     if( strcmp( name, ".." ) == 0 || strchr( name, '/' ) )
       { set_error_msg( "Directory access restricted" ); return false; }
     }
-  if( safe_names )
-    {
-    const char * p;
-    for( p = name; *p; ++p ) if( *p <= 31 && *p >= 1 )
-      { set_error_msg( "Control characters 1-31 not allowed in file names" );
+  const char * p;
+  for( p = name; *p; ++p )
+    if( *p == '\n' )
+      { set_error_msg( "Newline character not allowed in file names" );
         return false; }
-    }
+  if( safe_names )
+    for( p = name; *p; ++p )
+      if( ( *p <= 13 && *p >= 7 ) || *p == 27 || *p == 127 )
+        { set_error_msg( "Control characters not allowed in file names" );
+          return false; }
   return true;
   }
 
